package org.processmining.plugins.bpmn.analysis;

import java.awt.Color;
import java.awt.Dimension;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeTimestampImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.PNRepResultAllRequiredParamConnection;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNEdge;
import org.processmining.models.graphbased.directed.bpmn.BPMNNode;
import org.processmining.models.graphbased.directed.bpmn.elements.Activity;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.jgraph.ProMJGraphVisualizer;
import org.processmining.models.jgraph.visualization.ProMJGraphPanel;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.bpmn.analysis.pn.decorators.ActivityDecorator;
import org.processmining.plugins.converters.bpmn2pn.BPMN2PetriNetConverter;
import org.processmining.plugins.petrinet.replayer.PNLogReplayer;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.petrinet.replayresult.StepTypes;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;

import nl.tue.astar.AStarException;


@Plugin(name = "BPMN Model Analysis",
parameterLabels = { "Log", "BPMN Model" },
returnLabels = { "Annotated BPMN model"}, 
returnTypes = { BPMNDiagram.class}, 
userAccessible = true)

/**
 * 
 * @author Anna Kalenkova (akalenkova@hse.ru)
 *
 */
public class BPMNModelAnalysis {
	
	class LogModelMoves {
		public long logModelGood;
		public long logModelNoGood;
		public long logOnly;
		public long modelOnly;
	}
	
	@UITopiaVariant(uiLabel = "BPMN Model Analysis", affiliation = "HSE", author = "A. Kalenkova", email = "akalenkova@hse.ru")
	@PluginVariant(variantLabel = "BPMN Model Analysis", requiredParameterLabels = { 0, 1 })
	public BPMNDiagram bpmnModelAnalysis(final UIPluginContext context, XLog log, BPMNDiagram bpmnDiagram)  {
				
		// Retrieve Petri net
		BPMN2PetriNetConverter converter = new BPMN2PetriNetConverter(bpmnDiagram);
		converter.convert();
		Petrinet petrinet = converter.getPetriNet();
		
		// Replay log
		PNRepResult replayResult = null;
		try {
			replayResult = context.tryToFindOrConstructFirstObject(PNRepResult.class, 
					PNRepResultAllRequiredParamConnection.class, 
					PNRepResultAllRequiredParamConnection.PNREPRESULT, petrinet, log);
		} catch (ConnectionCannotBeObtained e) {
			e.printStackTrace();
			System.out.println("No connection found");
		}
		
		if(replayResult == null) {
			try {
				PNLogReplayer replayer = new PNLogReplayer();
				replayResult = replayer.replayLog(context, petrinet, log);
				
			} catch (ConnectionCannotBeObtained e) {
				e.printStackTrace();
			}  catch (AStarException e) {
				e.printStackTrace();
			}
		}
		
		annotate(context, bpmnDiagram, petrinet, converter, log, replayResult);

		return bpmnDiagram;
	}
	
	private BPMNDiagram annotate(UIPluginContext context, BPMNDiagram bpmnDiagram, Petrinet petriNet,
			BPMN2PetriNetConverter converter, XLog eventLog, PNRepResult replayResult) {
		Map<BPMNNode, Set<PetrinetNode>> transitionConversionMap = converter.getNodeMap();
		Map<BPMNEdge<BPMNNode, BPMNNode>, Place> placeConversionMap = converter.getFlowMap();
		Map<String, LogModelMoves> activitiesMoves = new HashMap<String, LogModelMoves>();
		Map<String, LogModelMoves> flowsMoves = new HashMap<String, LogModelMoves>();

		Map<Transition, Map<Integer,Long>> currentStartTime = new HashMap<Transition, Map<Integer,Long>>();
		Map<Transition, Set<Long>> executionTimes = new HashMap<Transition, Set<Long>>();
		
		// Calculate asynch moves
		Marking initMarking = new Marking();
		Marking finalMarking = new Marking();
		try {
		InitialMarkingConnection initialMarkingConnection = context.getConnectionManager()
				.getFirstConnection(InitialMarkingConnection.class, context, petriNet);
		initMarking = initialMarkingConnection.getObjectWithRole(InitialMarkingConnection.MARKING);
		
		FinalMarkingConnection finalMarkingConnection = context.getConnectionManager()
				.getFirstConnection(FinalMarkingConnection.class, context, petriNet);
		finalMarking = finalMarkingConnection.getObjectWithRole(FinalMarkingConnection.MARKING);
		
		} catch (ConnectionCannotBeObtained e) {
			e.printStackTrace();
		}
		
		Iterator<SyncReplayResult> iterator = replayResult.iterator();
		while (iterator.hasNext()) {
			SyncReplayResult res = iterator.next();
			Set<Integer> traces = res.getTraceIndex();
			Integer numOfTraces = traces.size();
			System.out.println(traces);
			int nodeNumber = 0;
			int eventNumber = 0;
			Marking currentMarking = new Marking(initMarking);
			for (StepTypes stepType : res.getStepTypes()) {
				Marking prevMarking = new Marking(currentMarking);
				Object node = res.getNodeInstance().get(nodeNumber);
				if (node instanceof Transition) {
					fireTransition((Transition) node, petriNet, currentMarking, eventNumber, traces, eventLog, 
							stepType, transitionConversionMap, bpmnDiagram, currentStartTime, executionTimes);
					Activity activity = retrieveActivityByTransition(transitionConversionMap, bpmnDiagram, (Transition) node);
					if (activity != null) {
						if (activitiesMoves.get(activity.getLabel()) == null) {
							activitiesMoves.put(activity.getLabel(), new LogModelMoves());
						}
						switch (stepType) {
							case LMGOOD :
								activitiesMoves.get(activity.getLabel()).logModelGood =
										activitiesMoves.get(activity.getLabel()).logModelGood + numOfTraces;
								break;
							case LMNOGOOD :
								activitiesMoves.get(activity.getLabel()).logModelNoGood =
										activitiesMoves.get(activity.getLabel()).logModelNoGood + numOfTraces;
								break;
							case L :
								activitiesMoves.get(activity.getLabel()).logOnly =
										activitiesMoves.get(activity.getLabel()).logOnly + numOfTraces;
								break;
							case MREAL :
								activitiesMoves.get(activity.getLabel()).modelOnly = 
										activitiesMoves.get(activity.getLabel()).modelOnly + numOfTraces;
								break;
							default :
								break;
						}
					}
				}
				for (Place place : prevMarking) {
					BPMNEdge flow = retrieveEdgeByPlace(placeConversionMap,  bpmnDiagram, place);
					if (flow != null) {
						if (flowsMoves.get(flow.getEdgeID().toString()) == null) {
							flowsMoves.put(flow.getEdgeID().toString(), new LogModelMoves());
						}
						switch (stepType) {
							case LMGOOD :
								flowsMoves.get(flow.getEdgeID().toString()).logModelGood = 
										flowsMoves.get(flow.getEdgeID().toString()).logModelGood + numOfTraces;
								break;
							case LMNOGOOD :
								flowsMoves.get(flow.getEdgeID().toString()).logModelNoGood =
										flowsMoves.get(flow.getEdgeID().toString()).logModelNoGood + numOfTraces;
								break;
							case L :
								flowsMoves.get(flow.getEdgeID().toString()).logOnly = 
										flowsMoves.get(flow.getEdgeID().toString()).logOnly + numOfTraces;
								break;
							case MREAL :
								flowsMoves.get(flow.getEdgeID().toString()).modelOnly = 
										flowsMoves.get(flow.getEdgeID().toString()).modelOnly + numOfTraces;
								break;
							default :
								break;
						}
					}
				}
				if(!stepType.equals(StepTypes.MREAL) && !stepType.equals(StepTypes.MINVI)) {
					eventNumber++;
				}
				nodeNumber++;
			}
		}
		// Find average execution times
		Map<Activity, Long> avgTimes = new HashMap<Activity, Long>();
		long maxTime = 0;
		for(Activity activity : bpmnDiagram.getActivities()) {
			Transition transition = (Transition)converter.getNodeMap().get(activity).iterator().next();
			Set<Long> allTimes = executionTimes.get(transition);
			if(allTimes != null) {
				long sum = 0; 
				for(Long time : allTimes) {
					sum+=time;
				}
				if(allTimes.size() != 0) {
					avgTimes.put(activity, sum/allTimes.size());
					if(maxTime < sum/allTimes.size()) {
						maxTime = sum/allTimes.size();
					}
				} else {
					avgTimes.put(activity,new Long(0));
				}
			}
		}
		ProMJGraphPanel proMJGraphPanel = ProMJGraphVisualizer.instance().visualizeGraphWithoutRememberingLayout(bpmnDiagram);
		//ViewSpecificAttributeMap viewSpecificMap = proMJGraphPanel.getGraph().getViewSpecificAttributes();
		// Enhance a diagram
		for (Activity activity : bpmnDiagram.getActivities()) {
			activity.getAttributeMap().put(AttributeMap.SIZE, new Dimension(100, 50));
			activity.getAttributeMap().put(AttributeMap.RESIZABLE, false);
			LogModelMoves activityMoves = activitiesMoves.get(activity.getLabel());
			if (activityMoves != null) {
				long logOnly = activityMoves.logOnly;
				long modelOnly = activityMoves.modelOnly;
				long logModel = activityMoves.logModelGood;
				long logModelNoGood = activityMoves.logModelNoGood;
				
				long total = logOnly + modelOnly + logModel + logModelNoGood;
				
				if(maxTime > 0) {
					activity.getAttributeMap().put(AttributeMap.FILLCOLOR, new Color(255, (int)(((1-(double)avgTimes.get(activity)/(double)maxTime)) * 255),
							(int)(((1-(double)avgTimes.get(activity)/(double)maxTime)) * 255)));
				}
				activity.setDecorator(new ActivityDecorator(logModel, total, avgTimes.get(activity), maxTime));
				proMJGraphPanel.getGraph().update(bpmnDiagram.getNodes());
			}
		}
		for (BPMNEdge edge : bpmnDiagram.getEdges()) {
			LogModelMoves flowMoves = flowsMoves.get(edge.getEdgeID().toString());
			if (flowMoves != null) {
				long logOnly = flowMoves.logOnly;
				long modelOnly = flowMoves.modelOnly;
				long logModel = flowMoves.logModelGood;
				long logModelNoGood = flowMoves.logModelNoGood;
				
				long total = logOnly + logModel;
				
				double onlyLogRatio = (double)logOnly / (double)total;
				
				edge.getAttributeMap().put(AttributeMap.EDGECOLOR, new Color ((int)(onlyLogRatio * 255),
						0, 0));
				edge.getAttributeMap().put(AttributeMap.LINEWIDTH, (float)2.5);
				
				edge.getAttributeMap().put(AttributeMap.LABELCOLOR, new Color ((int)(onlyLogRatio * 255),
						0, 0));
				
				edge.getAttributeMap().put(AttributeMap.LABEL, edge.getLabel() + " " + logOnly);
			}
		}
	return bpmnDiagram;
}

	private static void fireTransition(Transition transition, Petrinet petriNet, Marking marking, Integer eventNumber,
			Set<Integer> traces, XLog eventLog, StepTypes stepType,
			Map<BPMNNode, Set<PetrinetNode>> transitionConversionMap, BPMNDiagram diagram,
			Map<Transition, Map<Integer, Long>> currentStartTime, Map<Transition, Set<Long>> executionTimes) {
		for (PetrinetEdge edge : petriNet.getInEdges(transition)) {
			Place inPlace = (Place) edge.getSource();
			if (marking.contains(inPlace)) {
				marking.remove(inPlace);
			}
		}
		for (PetrinetEdge edge : petriNet.getOutEdges(transition)) {
			Place outPlace = (Place) edge.getTarget();
			marking.add(outPlace);
		}

		if (!stepType.equals(StepTypes.MREAL)) {
			for (Integer intTrace : traces) {
				long currentActivityTime = 0;
				if(!stepType.equals(StepTypes.MINVI)) {
				XTrace trace = eventLog.get(intTrace);
				XEvent event = trace.get(eventNumber);
					if ((XAttributeTimestampImpl) event.getAttributes().get("time:timestamp") != null) {
						currentActivityTime = ((XAttributeTimestampImpl) event.getAttributes().get("time:timestamp"))
								.getValueMillis();
						if (currentStartTime.get(transition) != null) {
							if (currentStartTime.get(transition).get(intTrace) != null) {
								long startTime = currentStartTime.get(transition).get(intTrace);
								if (executionTimes.get(transition) == null) {
									executionTimes.put(transition, new HashSet<Long>());
								}
								if(startTime != 0) {
									executionTimes.get(transition).add(currentActivityTime - startTime);
								}
								currentStartTime.get(transition).remove(intTrace);
							}
						}
					}
				} else {
					if (currentStartTime.get(transition) != null) {
						if (currentStartTime.get(transition).get(intTrace) != null) {
							long startTime = currentStartTime.get(transition).get(intTrace);
							currentActivityTime = startTime;
							if (executionTimes.get(transition) == null) {
								executionTimes.put(transition, new HashSet<Long>());
							}
							executionTimes.get(transition).add(new Long(0));
							currentStartTime.get(transition).remove(intTrace);
						}
					}
				}
				
				// Add new start times
				for (Transition newTransition : petriNet.getTransitions()) {
					boolean isEnabled = true;
					for (PetrinetEdge<?, ?> inEdge : petriNet.getInEdges(newTransition)) {
						if (!marking.contains(inEdge.getSource())) {
							isEnabled = false;
						}
					}
					if (isEnabled) {
						if (currentStartTime.get(newTransition) == null) {
							currentStartTime.put(newTransition, new HashMap<Integer, Long>());
						}
						if (currentStartTime.get(newTransition).get(intTrace) == null) {
							currentStartTime.get(newTransition).put(intTrace, currentActivityTime);
						}
					}
				}
			}
		}
	}
	
	private static Activity retrieveActivityByTransition(Map<BPMNNode, Set<PetrinetNode>> transitionConversionMap, BPMNDiagram diagram, Transition transition) {
		for (Activity activity : diagram.getActivities()) {
			if(transitionConversionMap.get(activity).contains(transition)) {
				return activity;
			}
		}	
		
		return null;
	}
	
	private static BPMNEdge<? extends BPMNNode, ? extends BPMNNode> retrieveEdgeByPlace(Map<BPMNEdge<BPMNNode, BPMNNode>, Place> placeConversionMap, 
			BPMNDiagram diagram, Place place) {
		for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> edge : diagram.getEdges()) {
			if(placeConversionMap.get(edge).equals(place)) {
				return edge;
			}
		}	
		
		return null;
	}
}
	

