package org.processmining.plugins.bpmn.analysis;

import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.PluginContextID;
import org.processmining.framework.plugin.PluginExecutionResult;
import org.processmining.framework.plugin.PluginParameterBinding;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.events.Logger;
import org.processmining.models.connections.flexiblemodel.FlexEndTaskNodeConnection;
import org.processmining.models.connections.flexiblemodel.FlexStartTaskNodeConnection;
import org.processmining.models.flexiblemodel.EndTaskNodesSet;
import org.processmining.models.flexiblemodel.Flex;
import org.processmining.models.flexiblemodel.FlexNode;
import org.processmining.models.flexiblemodel.StartTaskNodesSet;
import org.processmining.plugins.bpmn.analysis.settings.ModelToLogConnection;
import org.processmining.plugins.bpmn.analysis.settings.ui.ReplayerSettingsUI;
import org.processmining.plugins.flex.replayer.algorithms.ExtendedCostBasedAStarLogReplayAlgorithm;
import org.processmining.plugins.flex.replayer.algorithms.IFlexLogReplayAlgorithm;
import org.processmining.plugins.flex.replayer.ui.ParamSettingExtendedCostBasedReplay;
import org.processmining.plugins.flex.replayresult.performance.FlexPerfRepInfo;

public class CallReplayerHeadless {

	@Plugin(name = "(debug) call miner and replayer headless",
			parameterLabels = { "Log" },
			returnLabels = { "replay result" }, 
			returnTypes = { FlexPerfRepInfo.class }, 
			userAccessible = true, 
			help = "(debug) call miner and replayer headless")
	@UITopiaVariant(uiLabel = "call miner and replayer headless", 
					affiliation = UITopiaVariant.EHV, 
					author = "D. Fahland", email = "d.fahland@tue.nl")
	public FlexPerfRepInfo callMinerAndReplayerHeadless(final PluginContext context, XLog log) {

		System.out.println("mining flex model");
		Flex flex;
		try {
			flex = mineFlexModel(context, log);
		} catch (ConnectionCannotBeObtained e) {
			context.getFutureResult(0).cancel(true);
			System.out.println("failed to mined model:\n"+e);
			e.printStackTrace();
			return null;
		}
		
		System.out.println("loading default flex replayer configuration");
		Object[] flexReplayerConfiguration = new Object[3];

		ModelToLogConnection c = new ModelToLogConnection(flex, log);
		flexReplayerConfiguration[0] = c.getMapping();
		flexReplayerConfiguration[1] = new ExtendedCostBasedAStarLogReplayAlgorithm();
		ParamSettingExtendedCostBasedReplay p = new ParamSettingExtendedCostBasedReplay(false);
		flexReplayerConfiguration[2] = p.getAllParameters();

		return callReplayer(context, log, flex, flexReplayerConfiguration);
	}
			
	public Flex mineFlexModel(PluginContext context, XLog log) throws ConnectionCannotBeObtained {
		Flex flex = context.tryToFindOrConstructFirstObject(Flex.class, FlexStartTaskNodeConnection.class,
					FlexStartTaskNodeConnection.FLEX, log);
		return flex;
	}

	public FlexPerfRepInfo callReplayer(final PluginContext context, XLog log, Flex flex, Object[] flexReplayerConfiguration) {
		try {
			FlexStartTaskNodeConnection connStartTask = context.getConnectionManager()
					.getFirstConnection(FlexStartTaskNodeConnection.class, context, flex);
			StartTaskNodesSet startTaskNode = connStartTask
					.getObjectWithRole(FlexStartTaskNodeConnection.STARTTASKNODES);
	
			FlexEndTaskNodeConnection connEndTask = context.getConnectionManager()
					.getFirstConnection(FlexEndTaskNodeConnection.class, context, flex);
			EndTaskNodesSet endTaskNode = connEndTask
					.getObjectWithRole(FlexEndTaskNodeConnection.ENDTASKNODES);

			@SuppressWarnings("unchecked")
			Map<FlexNode, XEventClass> mappings = (Map<FlexNode, XEventClass>) flexReplayerConfiguration[ReplayerSettingsUI.MAPPING];
	
			IFlexLogReplayAlgorithm iFlexLogReplayAlgorithm = (IFlexLogReplayAlgorithm) flexReplayerConfiguration[ReplayerSettingsUI.ALGORITHM];
			Object[] parameters = (Object[]) flexReplayerConfiguration[ReplayerSettingsUI.PARAMETERS];
	
			// RUN REPLAY LOG ON C-NET FOR PERFORMANCE PLUG-IN FOR THE LOG
			System.out.println("REPLAYING All Traces for performance analysis.");
			context.log("REPLAYING All Traces for performance analysis.");
			/* call without UI */
			FlexPerfRepInfo flexPerfRepResult = runPerformanceReplayerWithChildContext(context, flex, log,
					mappings, startTaskNode, endTaskNode, iFlexLogReplayAlgorithm, parameters);
			System.out.println("finished replay");
			
			if (flexPerfRepResult == null) {
				context.getFutureResult(0).cancel(true);
			}
			
			return flexPerfRepResult;
		} catch (ConnectionCannotBeObtained e) {
			System.out.println("failed to get start and end task nodes:\n"+e);
			context.getFutureResult(0).cancel(true);
			e.printStackTrace();
			return null;
		}
	}
	
	private FlexPerfRepInfo runPerformanceReplayerWithChildContext(PluginContext context, Flex originalModel, XLog log,
			Map<FlexNode, XEventClass> flexNodeEventClassMap, StartTaskNodesSet startTaskNodesSet,
			EndTaskNodesSet endTaskNodesSet, IFlexLogReplayAlgorithm iFlexLogReplayAlgorithm, Object[] parameters) {

		FlexPerfRepInfo flexPerfRepResult4EachSyncReplayResult = null;
		//FIXME keep a (clean) list of currently running child contexts
		//FIXME for the progress listener cancellation: cycle through this list and send cancellation messages
		final PluginContext child = context.createChildContext("replay");
		context.getPluginLifeCycleEventListeners().firePluginCreated(child);
		child.getLoggingListeners().add(new Logger() {

			public void log(String message, PluginContextID contextID, MessageLevel messageLevel) {
				// ignore
			}

			public void log(Throwable t, PluginContextID contextID) {
				//ignore
				//context.log(t);
			}

		});

		//FIXME don't assume that there is a plugin (and what if there are more)

		Set<PluginParameterBinding> plugins = child.getPluginManager().getPluginsAcceptingOrdered(PluginContext.class,
				false, originalModel.getClass(), log.getClass(), flexNodeEventClassMap.getClass(),
				startTaskNodesSet.getClass(), endTaskNodesSet.getClass(), iFlexLogReplayAlgorithm.getClass(),
				parameters.getClass());

		//We need the second... (could be better but its a quick fix anywho)
		Iterator<PluginParameterBinding> pIt = plugins.iterator();
		PluginParameterBinding plugin = pIt.next();

		PluginExecutionResult result = plugin.invoke(child, originalModel, log, flexNodeEventClassMap,
				startTaskNodesSet, endTaskNodesSet, iFlexLogReplayAlgorithm, parameters);

		//FIXME correctly handle exceptions
		try {
			result.synchronize();
		} catch (CancellationException e) {
			System.out.println("result.synchronize cannot be accomplished for this run of performance algorithm!");
			System.out.println(e);
			e.printStackTrace();
			flexPerfRepResult4EachSyncReplayResult = null;
		} catch (ExecutionException e) {
			System.out.println("result.synchronize cannot be accomplished for this run of performance algorithm!");
			System.out.println(e);
			e.printStackTrace();
			flexPerfRepResult4EachSyncReplayResult = null;
		} catch (InterruptedException e) {
			System.out.println("result.synchronize cannot be accomplished for this run of performance algorithm!");
			System.out.println(e);
			e.printStackTrace();
			flexPerfRepResult4EachSyncReplayResult = null;
		} catch (Exception e) {
			System.out.println("result.synchronize cannot be accomplished for this run of performance algorithm!");
			System.out.println(e);
			e.printStackTrace();
			flexPerfRepResult4EachSyncReplayResult = null;
		}
		if (result != null) {
			if (result.getResult(0) != null) {
				if (result.getResult(0) instanceof FlexPerfRepInfo) {
					flexPerfRepResult4EachSyncReplayResult = result.getResult(0);
				} else {
					flexPerfRepResult4EachSyncReplayResult = null;
				}
			}
		}
		return flexPerfRepResult4EachSyncReplayResult;
	}
}
