package org.processmining.plugins.bpmn.analysis.swimlanes;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.JOptionPane;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.extension.std.XExtendedEvent;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagramFactory;
import org.processmining.models.graphbased.directed.bpmn.BPMNEdge;
import org.processmining.models.graphbased.directed.bpmn.BPMNNode;
import org.processmining.models.graphbased.directed.bpmn.elements.Activity;
import org.processmining.models.graphbased.directed.bpmn.elements.Event;
import org.processmining.models.graphbased.directed.bpmn.elements.Event.EventType;
import org.processmining.models.graphbased.directed.bpmn.elements.Flow;
import org.processmining.models.graphbased.directed.bpmn.elements.Gateway;
import org.processmining.models.graphbased.directed.bpmn.elements.Swimlane;
import org.processmining.models.graphbased.directed.bpmn.elements.SwimlaneType;
import org.processmining.plugins.bpmn.analysis.swimlanes.ui.SwimlaneSettingsGUI;

import cern.colt.matrix.DoubleFactory2D;
import cern.colt.matrix.DoubleMatrix2D;

@Plugin(name = "Swimlane Discovery Plug-in", parameterLabels = { "BPMN Diagram", "Event Log" }, returnLabels = { "BPMN Diagram with Swimlanes" }, returnTypes = { BPMNDiagram.class }, userAccessible = true, help = "Discovers a BPMN with Swimlane")
public class SwimlaneDiscovery {
	public static final String ARTIFICIAL_PROM = "Artificial (ProM)";
	
	private static final String DEFAULT_POOL_NAME = "Pool";

	@UITopiaVariant(affiliation = "Cordys", author = "I. Bayraktar", email = "i.bayraktar@cordys.com")
	@PluginVariant(variantLabel = "Swimlane Discovery Plug-in", requiredParameterLabels = { 0, 1 })
	public BPMNDiagram converter(UIPluginContext context, BPMNDiagram bpmnWithoutSwimlanes, XLog log) {

		if (!doesContainUserData(log)) {
			JOptionPane.showMessageDialog(null, "The Event Log contains no user data!",
					"Unsatisfied Swimlane Requirement", JOptionPane.ERROR_MESSAGE);
			context.getFutureResult(0).cancel(true);
			context.log("Unsatisfied Swimlane Requirement");

			return null;
		} else {
			SwimlaneSettingsGUI swimlaneSettingsGUI = new SwimlaneSettingsGUI();

			InteractionResult result = context.showConfiguration("Swimlane Threshold Settings", swimlaneSettingsGUI);
			if (result.equals(InteractionResult.CANCEL)) {
				context.getFutureResult(0).cancel(true);
			}

			int threshold = swimlaneSettingsGUI.getSliderValue();
			
			return convert(context, bpmnWithoutSwimlanes, log, threshold);
		}
	}
	
	public static final int DEFAULT_THRESHOLD = 10;
	
	@PluginVariant(variantLabel = "Swimlane Discovery Plug-in", requiredParameterLabels = { 0, 1 })
	public BPMNDiagram converter(PluginContext context, BPMNDiagram bpmnWithoutSwimlanes, XLog log) {
		
		if (!doesContainUserData(log)) {
			JOptionPane.showMessageDialog(null, "The Event Log contains no user data!",
					"Unsatisfied Swimlane Requirement", JOptionPane.ERROR_MESSAGE);
			context.getFutureResult(0).cancel(true);
			context.log("Unsatisfied Swimlane Requirement");

			return null;
		} else {
			return convert(context, bpmnWithoutSwimlanes, log, DEFAULT_THRESHOLD);			
		}
	}
	
	public BPMNDiagram convert(PluginContext context, BPMNDiagram bpmnWithoutSwimlanes, XLog log, int threshold) {

		Progress progress = context.getProgress();

		progress.setMinimum(0);
		progress.setMaximum(log.size() + 4);
		progress.setIndeterminate(false);

		BPMNDiagram bpmnWithSwimlanes = BPMNDiagramFactory.cloneBPMNDiagram(bpmnWithoutSwimlanes);

		/*
		 * Discard the old swimlanes and produce new ones
		 */
		bpmnWithSwimlanes.getSwimlanes().clear();

		Map<String, Map<String, Integer>> eventAndOriginatorListMap = getUserActivityPercentageMap(log/*
																									 * ,
																									 * true
																									 */);
		Map<String, List<String>> activityRemoveUserMap = new HashMap<String, List<String>>();
		for (String activity : eventAndOriginatorListMap.keySet()) {
			List<String> removeUserList = new ArrayList<String>();
			Map<String, Integer> userNumberOfExecutionMap = eventAndOriginatorListMap.get(activity);
			/*
			 * System.out.println("Activity : " + activity);
			 * System.out.println("Activity : " + activity);
			 */
			for (String user : userNumberOfExecutionMap.keySet()) {
				/*
				 * System.out.println("User : " + user + "\t Performance : "
				 * + userNumberOfExecutionMap.get(user).intValue());
				 */
				if (userNumberOfExecutionMap.get(user).intValue() < threshold) {
					removeUserList.add(user);
				}
			}
			activityRemoveUserMap.put(activity, removeUserList);
		}

		for (String activity : activityRemoveUserMap.keySet()) {
			Map<String, Integer> userNumberOfExecutionMap = eventAndOriginatorListMap.get(activity);
			for (String user : activityRemoveUserMap.get(activity)) {
				userNumberOfExecutionMap.remove(user);
			}
		}

		List<String> consideredActivities = new ArrayList<String>();
		List<String> sameSwimlanedActivities = new ArrayList<String>();

		List<String> activityList = new ArrayList<String>(eventAndOriginatorListMap.keySet());

		Swimlane pool = bpmnWithSwimlanes.addSwimlane(DEFAULT_POOL_NAME, null, SwimlaneType.POOL);
		
		for (int i = 0; i < activityList.size(); i++) {
			Map<String, Integer> userNumberOfExecutionMap = eventAndOriginatorListMap.get(activityList.get(i));
			Set<String> userSet = userNumberOfExecutionMap.keySet();

			if (!consideredActivities.contains(activityList.get(i))) {
				sameSwimlanedActivities.clear();
				consideredActivities.add(activityList.get(i));
				sameSwimlanedActivities.add(activityList.get(i));
				for (int j = i + 1; j < activityList.size(); j++) {
					Map<String, Integer> userNumberOfExecutionMap2 = eventAndOriginatorListMap.get(activityList
							.get(j));
					Set<String> userSet2 = userNumberOfExecutionMap2.keySet();
					if (userSet.equals(userSet2)) {
						consideredActivities.add(activityList.get(j));
						sameSwimlanedActivities.add(activityList.get(j));
					}
				}
				Swimlane swimlane = bpmnWithSwimlanes.addSwimlane(userSet.toString(), pool, SwimlaneType.LANE);
				List<Activity> activity2List = getActivityList(bpmnWithSwimlanes, sameSwimlanedActivities);
				for (Activity activity : activity2List) {
					swimlane.addChild(activity);
					activity.setParentSwimlane(swimlane);
				}
			}

		}

		for (Gateway gateway : bpmnWithSwimlanes.getGateways()) {
			Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> inedges = bpmnWithSwimlanes
					.getInEdges(gateway);
			if (inedges.size() > 1) {// join
				BPMNEdge<? extends BPMNNode, ? extends BPMNNode> outedge = bpmnWithSwimlanes.getOutEdges(gateway)
						.iterator().next();
				BPMNNode bpmnNode = outedge.getTarget();
				if (bpmnNode instanceof Gateway) {
					Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> inedges2 = bpmnWithSwimlanes
							.getInEdges(bpmnNode);
					if (inedges2.size() > 1) {// join
						BPMNEdge<? extends BPMNNode, ? extends BPMNNode> outedge2 = bpmnWithSwimlanes
								.getOutEdges(bpmnNode).iterator().next();
						BPMNNode bpmnNode2 = outedge2.getTarget();
						// cant be a gateway then so get the parent swimlane of the node possibly activity or event
						Swimlane s = bpmnNode2.getParentSwimlane();
						s.addChild(gateway);
						s.addChild(bpmnNode);
						gateway.setParentSwimlane(s);
						bpmnNode.setParentSwimlane(s);
					} else {// split
							// get one of the in edges in the join
						Swimlane s = inedges.iterator().next().getSource().getParentSwimlane();
						s.addChild(gateway);
						gateway.setParentSwimlane(s);
					}
				} else {// event or activity
					Swimlane s = bpmnNode.getParentSwimlane();
					if (s != null) {
						s.addChild(gateway);
						gateway.setParentSwimlane(s);
					}

				}
			} else {// split
				BPMNNode bpmnNode = inedges.iterator().next().getSource();
				if (bpmnNode instanceof Gateway) {
					Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> inedges2 = bpmnWithSwimlanes
							.getInEdges(bpmnNode);
					if (inedges2.size() > 1) {// join
						BPMNEdge<? extends BPMNNode, ? extends BPMNNode> outedge = bpmnWithSwimlanes
								.getOutEdges(gateway).iterator().next();
						gateway.setParentSwimlane(outedge.getTarget().getParentSwimlane());
					} else {// split
						BPMNNode bpmnNode2 = inedges2.iterator().next().getSource();
						Swimlane s = bpmnNode2.getParentSwimlane();
						s.addChild(gateway);
						s.addChild(bpmnNode);
						gateway.setParentSwimlane(s);
						bpmnNode.setParentSwimlane(s);
					}
				} else { // event or activity
					Swimlane s = bpmnNode.getParentSwimlane();
					if (s != null) {
						s.addChild(gateway);
						gateway.setParentSwimlane(s);
					}
				}

			}
		}

		for (Event event : bpmnWithSwimlanes.getEvents()) {
			if (event.getEventType().equals(EventType.START)) {
				Swimlane s = bpmnWithSwimlanes.getOutEdges(event).iterator().next().getTarget().getParentSwimlane();
				if (s != null) {
					s.addChild(event);
					event.setParentSwimlane(s);
				}
			} else {
				Swimlane s = bpmnWithSwimlanes.getInEdges(event).iterator().next().getSource().getParentSwimlane();
				if (s != null) {
					s.addChild(event);
					event.setParentSwimlane(s);
				}
			}
		}
		for (Flow flow : bpmnWithSwimlanes.getFlows()) {
			Swimlane swimlane1 = flow.getSource().getParentSwimlane();
			Swimlane swimlane2 = flow.getTarget().getParentSwimlane();
			if (swimlane1 != null && swimlane2 != null) {
				if (swimlane1.equals(swimlane2)) {
					flow.setParent(swimlane2);
				}
			}
		}
		/*
		 * Swimlane swimlane = bpmn.addSwimlane("Pool", null);
		 * 
		 * for (BPMNNode bpmnNode : bpmn.getNodes()) { if
		 * (bpmnNode.getParentSwimlane()==null) {
		 * bpmnNode.setParentSwimlane(swimlane); } }
		 */

		return bpmnWithSwimlanes;
	}

	/**
	 * @param bpmn
	 * @param sameSwimlanedActivities
	 * @return
	 */
	public List<Activity> getActivityList(BPMNDiagram bpmn, List<String> sameSwimlanedActivities) {
		List<Activity> activity2List = new ArrayList<Activity>();
		for (Activity activity : bpmn.getActivities()) {
			if (sameSwimlanedActivities.contains(activity.getLabel())) {
				activity2List.add(activity);
			}
		}
		return activity2List;
	}

	/**
	 * @param log
	 * @return
	 */
	private Map<String, Map<String, Integer>> getUserActivityPercentageMap(XLog log/*
																					 * ,
																					 * boolean
																					 * withTransition
																					 */) {
		XLogInfo summary = XLogInfoFactory.createLogInfo(log);
		List<String> originatorList = new ArrayList<String>();
		XEventClasses originators = summary.getResourceClasses();

		for (int k = 0; k < originators.getClasses().size(); k++) {
			originatorList.add(originators.getByIndex(k).toString());
		}
		List<String> activityNameList = new ArrayList<String>();
		DoubleMatrix2D otMatrix = null;
		/* if (!withTransition) { */
		XEventClasses activityNameClasses = summary.getNameClasses();
		for (int k = 0; k < activityNameClasses.getClasses().size(); k++) {
			activityNameList.add(activityNameClasses.getByIndex(k).toString());
		}
		int numOriginator = originatorList.size();

		otMatrix = DoubleFactory2D.sparse.make(numOriginator, activityNameList.size(), 0);

		for (XTrace trace : log) {
			for (int k = 0; k < trace.size(); k++) {
				XExtendedEvent xEvent = XExtendedEvent.wrap(trace.get(k));
				int row = originatorList.indexOf(xEvent.getResource());
				if (row == -1) {
					continue;
				}
				int column = activityNameList.indexOf(xEvent.getName());
				otMatrix.set(row, column, otMatrix.get(row, column) + 1.0);
			}
		}
		/*
		 * } else { XEventClasses activityEventClasses =
		 * summary.getEventClasses(); for (int k = 0; k <
		 * activityEventClasses.getClasses().size(); k++) {
		 * activityNameList.add(activityEventClasses.getByIndex(k).toString());
		 * } int numOriginator = originatorList.size();
		 * 
		 * otMatrix = DoubleFactory2D.sparse.make(numOriginator,
		 * activityNameList.size(), 0);
		 * 
		 * for (XTrace trace : log) { for (int k = 0; k < trace.size(); k++) {
		 * XExtendedEvent xEvent = XExtendedEvent.wrap(trace.get(k)); int row =
		 * originatorList.indexOf(xEvent.getResource()); if (row == -1) {
		 * continue; } int column = activityNameList.indexOf(xEvent.getName() +
		 * "+" + xEvent.getTransition()); otMatrix.set(row, column,
		 * otMatrix.get(row, column) + 1.0); } } }
		 */
		Map<String, Map<String, Integer>> eventAndOriginatorListMap = new HashMap<String, Map<String, Integer>>();

		Map<String, Map<String, Integer>> tempEventAndOriginatorListMap = new HashMap<String, Map<String, Integer>>();
		Map<String, Integer> eventExecutionNumber = new HashMap<String, Integer>();
		for (int i = 0; i < otMatrix.rows(); i++) {
			for (int k = 0; k < otMatrix.columns(); k++) {
				String user = originatorList.get(i);
				String activity = activityNameList.get(k);
				Integer occurence = (int) otMatrix.get(i, k);
				if (occurence != 0) {
					if (tempEventAndOriginatorListMap.containsKey(activity)) {
						Map<String, Integer> originatorsMap = tempEventAndOriginatorListMap.get(activity);
						originatorsMap.put(user, occurence);
						eventExecutionNumber.put(activity, occurence + eventExecutionNumber.get(activity));
					} else {
						eventExecutionNumber.put(activity, occurence);
						Map<String, Integer> originatorsMap = new HashMap<String, Integer>();
						originatorsMap.put(user, occurence);
						tempEventAndOriginatorListMap.put(activity, originatorsMap);
					}
				}

			}
		}
		for (String activity : tempEventAndOriginatorListMap.keySet()) {
			Integer totalNumberOfExecution = eventExecutionNumber.get(activity);
			Map<String, Integer> readUserAndNumberOfExecutionMap = tempEventAndOriginatorListMap.get(activity);
			Map<String, Integer> writeUserAndNumberOfExecutionMap = new HashMap<String, Integer>();

			for (String user : readUserAndNumberOfExecutionMap.keySet()) {
				writeUserAndNumberOfExecutionMap.put(user, readUserAndNumberOfExecutionMap.get(user).intValue() * 100
						/ totalNumberOfExecution.intValue());
			}
			eventAndOriginatorListMap.put(activity, writeUserAndNumberOfExecutionMap);
		}
		return eventAndOriginatorListMap;
	}

	private boolean doesContainUserData(XLog log) {
		XLogInfo summary = XLogInfoFactory.createLogInfo(log);

		//List<String> originatorList = new ArrayList<String>();
		XEventClasses originators = summary.getResourceClasses();
		/*
		 * for (int k = 0; k < originators.getClasses().size(); k++) {
		 * originatorList.add(originators.getByIndex(k).toString()); }
		 */
		boolean result = true;

		if (originators.size() == 0)
			result = false;
		else {
			List<XEventClass> originatorsXEventClasses = new ArrayList<XEventClass>(originators.getClasses());
			for (XEventClass xEventClass : originatorsXEventClasses) {
				if ("".equals(xEventClass.toString()))
					result = false;
				else if (ARTIFICIAL_PROM.equals(xEventClass.toString())) {
					result = false;
				} else {
					result = true;
				}
			}
		}
		//TODO: Extend this for role and organizational info later on
		return result;
	}
}
