/**
 * 
 */
package org.processmining.plugins.bpmn.converter;

import java.awt.Dimension;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.flexiblemodel.FlexEndTaskNodeConnection;
import org.processmining.models.connections.flexiblemodel.FlexStartTaskNodeConnection;
import org.processmining.models.flexiblemodel.EndTaskNodesSet;
import org.processmining.models.flexiblemodel.Flex;
import org.processmining.models.flexiblemodel.FlexEdge;
import org.processmining.models.flexiblemodel.FlexNode;
import org.processmining.models.flexiblemodel.SetFlex;
import org.processmining.models.flexiblemodel.StartTaskNodesSet;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagramFactory;
import org.processmining.models.graphbased.directed.bpmn.BPMNEdge;
import org.processmining.models.graphbased.directed.bpmn.BPMNNode;
import org.processmining.models.graphbased.directed.bpmn.elements.Activity;
import org.processmining.models.graphbased.directed.bpmn.elements.Event;
import org.processmining.models.graphbased.directed.bpmn.elements.Event.EventType;
import org.processmining.models.graphbased.directed.bpmn.elements.Event.EventUse;
import org.processmining.models.graphbased.directed.bpmn.elements.Flow;
import org.processmining.models.graphbased.directed.bpmn.elements.Gateway;
import org.processmining.models.graphbased.directed.bpmn.elements.Gateway.GatewayType;
import org.processmining.plugins.bpmn.converter.ui.CNET2BPMNConverterGUI;
import org.processmining.plugins.bpmn.converter.ui.CNetCheckGUI;

/**
 * @author ibayraktar
 * 
 */
@Plugin(name = "Convert C-Net to BPMN", parameterLabels = { "Causal Net" }, returnLabels = { "BPMN Diagram" }, returnTypes = { BPMNDiagram.class }, userAccessible = true, help = "Converts C-Net to BPMN")
public class CNET2BPMNConverter {

	private static final String EXCLUSIVE_GATEWAY = "EXCLUSIVE-GATEWAY";
	private static final String INCLUSIVE_GATEWAY = "INCLUSIVE_GATEWAY";
	private static final String PARALLEL_GATEWAY = "PARALLEL-GATEWAY";
	private static int aHeight = 80;
	private static int aWidth = 140;
	private static String JOIN = "JOIN";
	private static String SPLIT = "SPLIT";
	private static Map<String, Flow> flowMap = new HashMap<String, Flow>();
	
	@UITopiaVariant(affiliation = "Cordys", author = "I. Bayraktar", email = "i.bayraktar@cordys.com")
	@PluginVariant(variantLabel = "BPMN 2.0 Conversion", requiredParameterLabels = { 0 })
	public BPMNDiagram converter(UIPluginContext context, Flex model) {
		
		StartTaskNodesSet startTaskNodesSet = getStartTaskNodesSet(context, model);
		EndTaskNodesSet endTaskNodesSet = getEndTaskNodesSet(context, model);

		CNetCheckGUI checkGUI = new CNetCheckGUI(model, startTaskNodesSet, endTaskNodesSet);
		InteractionResult result = context.showConfiguration("Causal Net Check", checkGUI);

		if (!checkGUI.isModelValid()) {
			context.getFutureResult(0).cancel(true);
			context.log("Can't continue because C-Net check failed!");
			return null;
		}
		
		BPMNVendorSettings bpmnSettings = new BPMNVendorSettings();
		
		// create GUI to change BPMN vendor settings
		CNET2BPMNConverterGUI cnet2bpmnConverterGUI = new CNET2BPMNConverterGUI(bpmnSettings);

		result = context.showConfiguration("BPMN Vendor Settings", cnet2bpmnConverterGUI);
		if (result.equals(InteractionResult.CANCEL)) {
			context.getFutureResult(0).cancel(true);
		} else {
			cnet2bpmnConverterGUI.updateBPMNsettingsFromUI();
		}
		
		return convert(context, model, startTaskNodesSet, endTaskNodesSet, bpmnSettings);
	}
	
	@PluginVariant(variantLabel = "BPMN 2.0 Conversion", requiredParameterLabels = { 0 })
	public BPMNDiagram converter(PluginContext context, Flex model) {
		StartTaskNodesSet startTaskNodesSet = getStartTaskNodesSet(context, model);
		EndTaskNodesSet endTaskNodesSet = getEndTaskNodesSet(context, model);
		
		if (CNetFixModel.modelHasErrors(model, startTaskNodesSet, endTaskNodesSet, new LinkedList<String>(), new LinkedList<String>())) {
			boolean fixed = CNetFixModel.solveErrors(model, startTaskNodesSet, endTaskNodesSet, new LinkedList<String>());
			if (!fixed) {
				context.getFutureResult(0).cancel(true);
				context.log("Can't continue because C-Net check failed!");
				return null;
			}
		}
		
		BPMNVendorSettings bpmnSettings = new BPMNVendorSettings();
		return convert(context, model, startTaskNodesSet, endTaskNodesSet, bpmnSettings);
	}

	private StartTaskNodesSet getStartTaskNodesSet(PluginContext context, Flex model) {
		
		String message = "Get StartTaskNodesSet of the given C-Net";
		try {
			context.getProgress().setCaption(message);
			context.getProgress().setValue(100);
			StartTaskNodesSet startTaskNodesSet = context.tryToFindOrConstructFirstObject(StartTaskNodesSet.class,
					FlexStartTaskNodeConnection.class, FlexStartTaskNodeConnection.STARTTASKNODES, model);
			return startTaskNodesSet;

		} catch (ConnectionCannotBeObtained e1) {
			context.getFutureResult(0).cancel(true);
			context.log("Can't obtain connection for " + message);
			return null;
		}
	}
	
	private EndTaskNodesSet getEndTaskNodesSet(PluginContext context, Flex model) {
		
		String message = "Get EndTaskNodesSet of the given C-Net";
		try {
			context.getProgress().setCaption(message);
			context.getProgress().setValue(100);

			EndTaskNodesSet endTaskNodesSet = context.tryToFindOrConstructFirstObject(EndTaskNodesSet.class,
					FlexEndTaskNodeConnection.class, FlexEndTaskNodeConnection.ENDTASKNODES, model);
			return endTaskNodesSet;

		} catch (ConnectionCannotBeObtained e1) {
			context.getFutureResult(0).cancel(true);
			context.log("Can't obtain connection for " + message);
			return null;
		}
	}

	public BPMNDiagram convert(PluginContext context, Flex model, StartTaskNodesSet startTaskNodesSet, EndTaskNodesSet endTaskNodesSet, BPMNVendorSettings bpmnSettings) {

		Progress progress = context.getProgress();

		/*
		 * Initialize the BPMN diagram.
		 */
		BPMNDiagram bpmn = BPMNDiagramFactory.newBPMNDiagram("BPMN Diagram");
		Map<String, BPMNNode> id2node = new HashMap<String, BPMNNode>();

		progress.setCaption("Converting To BPMN");
		progress.setIndeterminate(false);

		/*
		 * Convert C-Net to BPMN.
		 */
		convertAllActivities(bpmn, id2node, model/*, isFlexModelWithLifecycle(flex)*/);
		createStartEventAndConnectToStartActivity(model, bpmn, id2node, startTaskNodesSet);
		createEndEventAndConnectToEndActivity(model, bpmn, id2node, endTaskNodesSet);
		connectSequencesAndSplitGateways(bpmn, bpmnSettings, id2node, model);
		fillFlowMap(bpmn);//for remove purposes.
		connectJoinGateways(bpmn, bpmnSettings, id2node, model);
		setGatewayMarkingVisibilities(bpmn, bpmnSettings);

		System.out.println("Conversion Completed.");
		//print(bpmn);
		/*
		 * Set the label of the BPMN diagram.
		 */
		progress.setCaption("Getting BPMN Visualization");
		context.getFutureResult(0).setLabel("Converted BPMN Model");
		return bpmn;

	}

	/*private boolean isFlexModelWithLifecycle(Flex flex) {
		boolean b = false;
		Set<String> flexNodeStringSet = new HashSet<String>();
		Set<FlexNode> nodes = flex.getNodes();
		for (FlexNode flexNode : nodes) {
			if (flexNode.getLabel().indexOf("+") != -1) {
				String name = flexNode.getLabel().substring(0, flexNode.getLabel().indexOf("+"));
				flexNodeStringSet.add(name);
			}
		}
		if (flexNodeStringSet.size() == nodes.size())
			b = false;
		else
			b = true;
		return b;
	}*/

	private void setGatewayMarkingVisibilities(BPMNDiagram bpmn, BPMNVendorSettings settings) {
		for (Gateway gateway : bpmn.getGateways()) {
			if (gateway.getGatewayType().equals(GatewayType.DATABASED)) {
				gateway.setMarkerVisible(settings.getXORMarker());
			}
			if (gateway.getGatewayType().equals(GatewayType.INCLUSIVE)) {
				gateway.setMarkerVisible(settings.getORMarker());
			}
			if (gateway.getGatewayType().equals(GatewayType.PARALLEL)) {
				gateway.setMarkerVisible(settings.getANDMarker());
			}
		}
	}

	/**
	 * @param ahn
	 * @param bpmn
	 * @param id2node
	 */
	private static void createEndEventAndConnectToEndActivity(Flex flex, BPMNDiagram bpmn,
			Map<String, BPMNNode> id2node, EndTaskNodesSet flexEndTaskNodes) {

		for (SetFlex nodes : flexEndTaskNodes) {
			
			Event endEvent = bpmn.addEvent("END EVENT", EventType.END, null, EventUse.THROW, null);
			endEvent.getAttributeMap().put(AttributeMap.SIZE, new Dimension(aHeight, aHeight));
			
			BPMNNode connectTo = endEvent;
			
			if (nodes.size() > 1) {
				Gateway branchJoin = bpmn.addGateway(PARALLEL_GATEWAY, getGatewayType(PARALLEL_GATEWAY));
				bpmn.addFlow(endEvent, branchJoin, null);
				connectTo = branchJoin;
			}
			
			for (FlexNode node : nodes) {
				BPMNNode fromNode = id2node.get("ID" + node.getId());
				bpmn.addFlow(fromNode, connectTo, null);
			}
		}
	}

	/**
	 * @param ahn
	 * @param bpmn
	 * @param id2node
	 */
	private static void createStartEventAndConnectToStartActivity(Flex flex, BPMNDiagram bpmn,
			Map<String, BPMNNode> id2node, StartTaskNodesSet flexStartTaskNodes) {

		for (SetFlex nodes : flexStartTaskNodes) {

			Event startEvent = bpmn.addEvent("START EVENT", EventType.START, null, EventUse.CATCH, null);
			startEvent.getAttributeMap().put(AttributeMap.SIZE, new Dimension(aHeight, aHeight));
			
			BPMNNode connectTo = startEvent;
			
			if (nodes.size() > 1) {
				Gateway branchSplit = bpmn.addGateway(PARALLEL_GATEWAY, getGatewayType(PARALLEL_GATEWAY));
				bpmn.addFlow(startEvent, branchSplit, null);
				connectTo = branchSplit;
			}
			
			for (FlexNode node : nodes) {
				BPMNNode fromNode = id2node.get("ID" + node.getId());
				bpmn.addFlow(connectTo, fromNode, null);
			}
		}
	}

	/**
	 * @param bpmn
	 * @param id2node
	 * @param keys
	 */
	private static void convertAllActivities(BPMNDiagram bpmn, Map<String, BPMNNode> id2node, Flex flex/*,
			boolean isFlexModelWithLifecycle*/) {
		for (FlexNode entry : flex.getNodes()) {
			String name = entry.getLabel();
			/*if (!isFlexModelWithLifecycle) {*/
				if (entry.getLabel().indexOf("+") != -1) {
					name = entry.getLabel().substring(0, entry.getLabel().indexOf("+"));

				}
			/*}*/
			Activity a = bpmn.addActivity(name, false, false, false, false, false);
			a.getAttributeMap().put(AttributeMap.SIZE, new Dimension(aWidth, aHeight));
			id2node.put("ID" + entry.getId(), a);
		}
	}

	/**
	 * @param ahn
	 * @param bpmn
	 * @param id2node
	 * @param keys
	 */
	private void connectSequencesAndSplitGateways(BPMNDiagram bpmn, BPMNVendorSettings settings, Map<String, BPMNNode> id2node, Flex flex) {

		for (FlexNode entry : flex.getNodes()) {
			List<FlexEdge<? extends FlexNode, ? extends FlexNode>> outEdgesList = new ArrayList<FlexEdge<? extends FlexNode, ? extends FlexNode>>(
					flex.getOutEdges(entry));
			if (outEdgesList.size() < 1) {
				continue;
			} else if (outEdgesList.size() == 1) {
				FlexNode target = outEdgesList.get(0).getTarget();
				BPMNNode fromNode = id2node.get("ID" + entry.getId());
				BPMNNode toNode = id2node.get("ID" + target.getId());
				bpmn.addFlow(fromNode, toNode, null);
			} else {
				/*
				 * No Pattern exists if patternType is 0, but and-join/split if
				 * it is 1, then AND, 2 for XOR, 3 for OR
				 */
				PatternType patternType = identifyPatternType(SPLIT, entry);
				if (patternType == PatternType.AND) {
					fillANDORSplitGatewaySettings(flex, entry, PARALLEL_GATEWAY, id2node, bpmn,
							settings.getANDSplit().equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit));
				}
				else if (patternType == PatternType.XOR) {
					fillXORSplitGatewaySettings(bpmn, settings, id2node, flex, entry,
							settings.getXORSplit().equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit));
				} else {
					fillANDORSplitGatewaySettings(flex, entry, INCLUSIVE_GATEWAY, id2node, bpmn,
							settings.getORSplit().equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit));
				}
			}
		}
	}

	private void fillXORSplitGatewaySettings(BPMNDiagram bpmn, BPMNVendorSettings settings, Map<String, BPMNNode> id2node, Flex flex,
			FlexNode entry, boolean isExplicit) {
		// We have an XOR gateway with distinct output sets : {{A, B}, {C, D}, {E, F}} or distinct output sets : {{A}, {C, D}}
		// if one of the distinct output set of {{A}, {C, D}} has two or more nodes i.e. {C, D} to represent this we first need to make an XOR gateway
		// then one AND Gateway for each distinct set with at least two nodes.
		// if the none of the distinct sets:{{A}, {C}, {D}} have two or more nodes, then we will just create XOR gateway.

		// get out target nodes and create id for XOR gateway
		List<FlexNode> targetNodeList = getFlexNodeList(flex, entry, true);
		String id = "" + entry.getId();
		if (isExplicit) {
			for (FlexNode flexNode : targetNodeList) {
				id += flexNode.getId();
			}
			// CREATE GATEWAY ACTIVITY
			GatewayType type = getGatewayType(EXCLUSIVE_GATEWAY);
			Gateway g = bpmn.addGateway(EXCLUSIVE_GATEWAY, type);
			id2node.put("ID" + id, g);
			g.getAttributeMap().put(AttributeMap.SIZE,
					new Dimension(((aHeight + aWidth) / 2), ((aHeight + aWidth) / 2)));

			// CREATE NEW TRANSITION TO CONNECT entry ACTIVITy TO GATEWAY
			BPMNNode fromNode = id2node.get("ID" + entry.getId());
			BPMNNode toNode = id2node.get("ID" + id);
			bpmn.addFlow(fromNode, toNode, null);
		}

		Set<SetFlex> setFlexSet = entry.getOutputNodes();
		for (SetFlex setFlex : setFlexSet) {
			if (setFlex.size() == 1) {
				// CREATE NEW TRANSITION TO CONNECT NODE ACTIVITy IN OUTPUT SET TO GATEWAY
				BPMNNode tempFromNode = id2node.get("ID" + id);
				BPMNNode tempToNode = id2node.get("ID" + setFlex.first().getId());
				bpmn.addFlow(tempFromNode, tempToNode, null);
			} else {
				// THERE ARE TWO NODES IN OUTPUTSET SO:
				// 1. CREATE AND GATEWAY
				// 2. CREATE NEW TRANSITION TO CONNECT AND GATEWAY TO XOR GATEWAY
				// 3. CREATE NEW TRANSITIONS TO CONNECT NODES IN OUTPUTSET TO AND GATEWAY

				String id2 = "" + id;
				for (FlexNode tempNode : setFlex) {
					// CREATE ID FOR AND GATEWAY ACTIVITY
					id2 += tempNode.getId();
				}
				if (settings.getANDSplit()
						.equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit)) {
					// 1. CREATE AND GATEWAY
					GatewayType type2 = getGatewayType(PARALLEL_GATEWAY);
					Gateway g2 = bpmn.addGateway(PARALLEL_GATEWAY, type2);
					id2node.put("ID" + id2, g2);
					g2.getAttributeMap().put(AttributeMap.SIZE,
							new Dimension(((aHeight + aWidth) / 2), ((aHeight + aWidth) / 2)));
				} else {
					// 1. WE NEED TO CREATE A DUMMY ACTIVITY BECAUSE AND GATEWAY IS IMPLICIT AND WE NEED TO DISTINGUISH THIS SET FROM OTHER SETS
					Activity a = bpmn.addActivity("DUMMY", false, false, false, false, false);
					a.getAttributeMap().put(AttributeMap.SIZE, new Dimension(aWidth / 2, aHeight / 2));
					id2node.put("ID" + id2, a);
				}
				// 2. CREATE NEW TRANSITION TO CONNECT AND GATEWAY TO XOR GATEWAY
				BPMNNode tempFromNode = id2node.get("ID" + id);
				BPMNNode tempToNode = id2node.get("ID" + id2);
				bpmn.addFlow(tempFromNode, tempToNode, null);

				for (FlexNode tempNode : setFlex) {
					// 3. CREATE NEW TRANSITIONS TO CONNECT NODES IN OUTPUTSET TO AND GATEWAY
					BPMNNode tempFromNode2 = id2node.get("ID" + id2);
					BPMNNode tempToNode2 = id2node.get("ID" + tempNode.getId());
					bpmn.addFlow(tempFromNode2, tempToNode2, null);
				}
			}
		}
	}

	/**
	 * @param isOut
	 *            : false for Inedges, true for Outedges
	 * @return List<FlexNode>
	 */
	private List<FlexNode> getFlexNodeList(Flex flex, FlexNode entry, boolean isOut) {
		List<FlexNode> targetNodeList = new ArrayList<FlexNode>();
		Collection<FlexEdge<? extends FlexNode, ? extends FlexNode>> flexEdgesCollection = null;
		if (isOut) {
			flexEdgesCollection = flex.getOutEdges(entry);
			for (FlexEdge<? extends FlexNode, ? extends FlexNode> flexEdge : flexEdgesCollection) {
				targetNodeList.add(flexEdge.getTarget());
			}
		} else {
			flexEdgesCollection = flex.getInEdges(entry);
			for (FlexEdge<? extends FlexNode, ? extends FlexNode> flexEdge : flexEdgesCollection) {
				targetNodeList.add(flexEdge.getSource());
			}
		}

		return targetNodeList;
	}

	/**
	 * @param ahn
	 * @param bpmn
	 * @param id2node
	 * @param keys
	 */
	private void connectJoinGateways(BPMNDiagram bpmn, BPMNVendorSettings settings, Map<String, BPMNNode> id2node, Flex flex) {
		/*
		 * But we connected the graphs according to just splits however we need
		 * to add joins but while adding joins we need to remove the transitions
		 * we made then we will create the gateway and then connect the events
		 * to them lets do it
		 */
		// JOINS
		// FILL TRANSITION VARIABLES BUT REMOVE TRANSITIONS IN
		// WHICH JOIN EXISTS AND RECONNECT TO THE JOIN GATEWAY

		for (FlexNode entry : flex.getNodes()) {
			List<FlexEdge<? extends FlexNode, ? extends FlexNode>> inEdgesList = new ArrayList<FlexEdge<? extends FlexNode, ? extends FlexNode>>(
					flex.getInEdges(entry));
			if (inEdgesList.size() > 1) {
				PatternType patternType = identifyPatternType(JOIN, entry);

				if (patternType == PatternType.AND) {
					if (settings.getANDJoin()
							.equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit)) {
						fillJoinGatewaySettings(flex, settings, entry, PARALLEL_GATEWAY, id2node, bpmn);
					}
				} else if (patternType == PatternType.XOR) {
					if (settings.getXORJoin()
							.equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit)) {
						fillJoinGatewaySettings(flex, settings, entry, EXCLUSIVE_GATEWAY, id2node, bpmn);
					}

				} else if (patternType == PatternType.OR) {
					if (settings.getORJoin()
							.equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit)) {
						fillJoinGatewaySettings(flex, settings, entry, INCLUSIVE_GATEWAY, id2node, bpmn);
					}

				}
			}

		}
	}

	private void fillANDORSplitGatewaySettings(Flex flex, FlexNode entry, String activityNameStr,
			Map<String, BPMNNode> id2node, BPMNDiagram bpmn, boolean isExplicit) {
		List<FlexNode> targetNodeList = getFlexNodeList(flex, entry, true);
		String id = "" + entry.getId();
		if (isExplicit) {
			for (FlexNode flexNode : targetNodeList) {
				id += flexNode.getId();
			}
			// CREATE GATEWAY ACTIVITY
			GatewayType type = getGatewayType(activityNameStr);
			Gateway g = bpmn.addGateway(activityNameStr, type);
			id2node.put("ID" + id, g);
			g.getAttributeMap().put(AttributeMap.SIZE,
					new Dimension(((aHeight + aWidth) / 2), ((aHeight + aWidth) / 2)));
			// CREATE NEW TRANSITIONS TO CONNECT ACTIVITIES TO GATEWAY
			BPMNNode fromNode = id2node.get("ID" + entry.getId());
			BPMNNode toNode = id2node.get("ID" + id);
			bpmn.addFlow(fromNode, toNode, null);
		}
		for (FlexNode flexNode : targetNodeList) {

			BPMNNode tempFromNode = id2node.get("ID" + id);
			BPMNNode tempToNode = id2node.get("ID" + flexNode.getId());
			bpmn.addFlow(tempFromNode, tempToNode, null);
		}

	}

	/**
	 * @param ahn
	 * @param index
	 * @param patternType
	 * @param patternTypeCode
	 * @param activityNameStr
	 * @param gatewayTypeStr
	 * @param splitJoinStr
	 */
	private void fillJoinGatewaySettings(Flex flex, BPMNVendorSettings settings, FlexNode entry, String activityNameStr,
			Map<String, BPMNNode> id2node, BPMNDiagram bpmn) {
		// CREATE ID FOR GATEWAY ACTIVITY
		String id = "" + entry.getId();
		List<FlexNode> sourceNodeList = getFlexNodeList(flex, entry, false);
		for (FlexNode node : sourceNodeList) {
			id += node.getId();
		}
		// CREATE GATEWAY ACTIVITY

		String name = activityNameStr;
		GatewayType type = getGatewayType(activityNameStr);
		Gateway g = bpmn.addGateway(name, type);
		g.getAttributeMap().put(AttributeMap.SIZE, new Dimension(((aHeight + aWidth) / 2), ((aHeight + aWidth) / 2)));
		id2node.put("ID" + id, g);

		// REMOVE TRANSITIONS COMING FROM THE SPLIT WHICH CONNECTS DIRECTLY
		// EVENTS TOGETHER
		// AND PUT A GATEWAY BETWEEN THESE ACTIVITIES AND CONNECT THEM
		for (FlexNode node : sourceNodeList) {
			// IF THERE IS XOR/AND/OR SPLIT GATEWAY BETWEEN patternEvent AND
			// index EVENT IDS, WE NEED TO REMOVE THE TRANSITION BETWEEN GATEWAY
			// AND index
			//print(bpmn);
			List<FlexEdge<? extends FlexNode, ? extends FlexNode>> outEdgesList = new ArrayList<FlexEdge<? extends FlexNode, ? extends FlexNode>>(
					flex.getOutEdges(node));
			if (outEdgesList.size() > 1) {
				PatternType patternType = identifyPatternType(SPLIT, node);
				List<FlexNode> targetNodeList = getFlexNodeList(flex, node, true);
				String tempID = "" + node.getId();
				for (FlexNode tempNode : targetNodeList) {
					tempID += tempNode.getId();
				}
				// CHECK IF THE VENDOR SPECIFIC NOTATION HAS EXPLICIT GATEWAY
				// FOR THE SPECIFIC SPLIT
				if ((patternType == PatternType.AND && settings.getANDSplit()
						.equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit))
						|| (patternType == PatternType.OR && settings.getORSplit()
								.equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit))) {
					removeOldAndConnectNewTransitions2Gateway("" + entry.getId(), id, tempID, id2node, bpmn);

				} else if (patternType == PatternType.XOR
						&& settings.getXORSplit()
								.equals(BPMNVendorSettings.GatewayVisibilitySettings.Explicit)) {
					removeOldAndConnectNewTransitions2XORGateway(entry, id2node, bpmn, id, node, tempID);
				} else {
					// THE VENDOR SPECIFIC NOTATION HAS NO EXPLICIT GATEWAY FOR THE SPECIFIC SPLIT
					removeOldAndConnectNewTransitions2Gateway("" + entry.getId(), id, "" + node.getId(), id2node, bpmn);
				}
			} else {
				// IF THERE IS NO XOR/AND/OR SPLIT GATEWAY BETWEEN patternEvent
				// AND index EVENT IDS, JUST REMOVE THE TRANSITION BETWEEN THEM
				removeOldAndConnectNewTransitions2Gateway("" + entry.getId(), id, "" + node.getId(), id2node, bpmn);
			}
			//print(bpmn);
		}

		// CREATE NEW TRANSITION WHICH CONNECTS THE GATEWAY TO THE INDEX ELEMENT
		BPMNNode fromNode = id2node.get("ID" + id);
		BPMNNode toNode = id2node.get("ID" + entry.getId());
		bpmn.addFlow(fromNode, toNode, null);

	}

	private void removeOldAndConnectNewTransitions2XORGateway(FlexNode entry, Map<String, BPMNNode> id2node,
			BPMNDiagram bpmn, String id, FlexNode node, String tempID) {
		Set<SetFlex> setFlexSet = node.getOutputNodes();
		for (SetFlex setFlex : setFlexSet) {
			if (setFlex.size() == 1 && setFlex.first().equals(entry)) {
				removeOldAndConnectNewTransitions2Gateway("" + entry.getId(), id, tempID, id2node, bpmn);
			} else if (setFlex.size() >= 2 && setFlex.contains(entry)) {
				// THERE ARE TWO NODES IN OUTPUTSET SO:
				// 1. CREATE AND GATEWAY
				// 2. CREATE NEW TRANSITION TO CONNECT AND GATEWAY TO XOR GATEWAY
				// 3. CREATE NEW TRANSITIONS TO CONNECT NODES IN OUTPUTSET TO AND GATEWAY

				String id2 = "" + tempID;
				for (FlexNode tempNode : setFlex) {
					// CREATE ID FOR AND GATEWAY ACTIVITY
					id2 += tempNode.getId();
				}
				removeOldAndConnectNewTransitions2Gateway("" + entry.getId(), id, id2, id2node, bpmn);
			}
		}
	}

	/**
	 * @param activityNameStr
	 * @return
	 */
	private static GatewayType getGatewayType(String activityNameStr) {
		GatewayType type = GatewayType.DATABASED;

		if (activityNameStr.equalsIgnoreCase(EXCLUSIVE_GATEWAY)) {
			type = GatewayType.DATABASED;
		} else if (activityNameStr.equalsIgnoreCase(PARALLEL_GATEWAY)) {
			type = GatewayType.PARALLEL;
		} else if (activityNameStr.equalsIgnoreCase(INCLUSIVE_GATEWAY)) {
			type = GatewayType.INCLUSIVE;
		}
		return type;
	}

	/**
	 * @param index
	 * @param id
	 * @param tempID
	 */
	private static void removeOldAndConnectNewTransitions2Gateway(String index, String id, String tempID,
			Map<String, BPMNNode> id2node, BPMNDiagram bpmn) {
		BPMNNode fromNode = id2node.get("ID" + tempID);
		BPMNNode toNode = id2node.get("ID" + index);
		Flow f = flowMap.get(fromNode.getId() + "-" + toNode.getId());
		bpmn.removeEdge(f);

		BPMNNode fromNode2 = id2node.get("ID" + tempID);
		BPMNNode toNode2 = id2node.get("ID" + id);
		bpmn.addFlow(fromNode2, toNode2, null);
	}

	/**
	 * @param bpmn
	 */
	public static void fillFlowMap(BPMNDiagram bpmn) {
		flowMap.clear();
		for (Flow flow : bpmn.getFlows()) {
			if (!flowMap.containsKey(flow.getSource() + "-" + flow.getTarget()))
				flowMap.put(flow.getSource().getId() + "-" + flow.getTarget().getId(), flow);
			/*
			 * if (flowMap.containsKey(flow.getSource() + "-" +
			 * flow.getTarget()))
			 * 
			 * System.out.println(""); else { System.out.println("Flow ID: " +
			 * flow.hashCode() + "Flow Source: " + flow.getSource().getLabel() +
			 * "Flow Source ID: " + flow.getSource().getId() + "Flow Target: " +
			 * flow.getTarget().getLabel() + "Flow Target ID: " +
			 * flow.getTarget().getId()); flowMap.put(flow.getSource().getId() +
			 * "-" + flow.getTarget().getId(), flow); }
			 */
		}
	}

	public static enum PatternType {
		AND,
		XOR,
		OR
	};
	
	/**
	 * @return int 1 for AND, 2 for XOR, 3 for OR
	 */
	public PatternType identifyPatternType(String joinSplit, FlexNode entry) {

		Set<SetFlex> setFlexSet = null;
		if (joinSplit.equals(JOIN))
			setFlexSet = entry.getInputNodes();
		else
			setFlexSet = entry.getOutputNodes();

		int size = setFlexSet.size();

		if (size == 1)
			return PatternType.AND;
		else if (hasXOR(setFlexSet))
			return PatternType.XOR;
		else
			return PatternType.OR;
	}

	private boolean hasXOR(Set<SetFlex> setFlexSet) {
		// check whether all SetFlex in setFlexSet are distinct: if it is return
		// true, because it is an XOR, otherwise it is OR so return false
		boolean b = false;
		for (SetFlex setFlex : setFlexSet) {
			for (SetFlex setFlex2 : setFlexSet) {
				SetFlex newTempSetFlex = (SetFlex) setFlex.clone();
				SetFlex newTempSetFlex2 = (SetFlex) setFlex2.clone();
				if (!newTempSetFlex.equals(newTempSetFlex2)) {
					newTempSetFlex.retainAll(newTempSetFlex2);
					if (newTempSetFlex.isEmpty())
						continue;
					else {
						b = true;
						break;
					}
				}

			}
			if (b)
				break;// not an XOR
		}
		return !b;
	}

	public void test(TreeSet<Integer> eventIndicesTreeSet, Integer eventIndex) {

		if (eventIndex == eventIndicesTreeSet.size())
			return;
		else {
			NavigableSet<Integer> eventIndicesNavigableSet = eventIndicesTreeSet.tailSet(eventIndex, false);

			test(eventIndicesTreeSet, eventIndex + 1);
			test2(eventIndicesNavigableSet, eventIndex + 1);
		}

	}

	public void test2(NavigableSet<Integer> eventIndicesTreeSet, Integer eventIndex) {

	}

	public static <T> List<TreeSet<T>> powerSet(Set<T> set) {

		List<TreeSet<T>> listOfSubsets = new ArrayList<TreeSet<T>>();
		if (set.isEmpty()) {
			listOfSubsets.add(new TreeSet<T>());
			return listOfSubsets;
		}

		List<T> list = new ArrayList<T>(set);
		T head = list.get(0);
		TreeSet<T> rest = new TreeSet<T>(list.subList(1, list.size()));

		for (TreeSet<T> restSubsets : powerSet(rest)) {
			TreeSet<T> newSet = new TreeSet<T>();
			newSet.add(head);
			newSet.addAll(restSubsets);
			listOfSubsets.add(newSet);
			listOfSubsets.add(restSubsets);
		}
		return listOfSubsets;
	}

	public void print(BPMNDiagram BPMNWithoutSwimlanes) {

		for (BPMNNode bpmnNode : BPMNWithoutSwimlanes.getNodes()) {
			System.out.println("Node:" + bpmnNode);
			Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> bpmnEdges = BPMNWithoutSwimlanes
					.getInEdges(bpmnNode);
			System.out.println("In:");
			for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> bpmnEdge : bpmnEdges) {
				System.out.println("Target: " + bpmnEdge.getTarget());
				System.out.println("Source: " + bpmnEdge.getSource());

			}
			Collection<BPMNEdge<? extends BPMNNode, ? extends BPMNNode>> bpmnEdges2 = BPMNWithoutSwimlanes
					.getOutEdges(bpmnNode);
			System.out.println("Out:");
			for (BPMNEdge<? extends BPMNNode, ? extends BPMNNode> bpmnEdge : bpmnEdges2) {
				System.out.println("Target: " + bpmnEdge.getTarget());
				System.out.println("Source: " + bpmnEdge.getSource());

			}

		}
	}
}
