package org.processmining.plugins.bpmn.converter;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.processmining.models.flexiblemodel.EndTaskNodesSet;
import org.processmining.models.flexiblemodel.Flex;
import org.processmining.models.flexiblemodel.FlexEdge;
import org.processmining.models.flexiblemodel.FlexNode;
import org.processmining.models.flexiblemodel.SetFlex;
import org.processmining.models.flexiblemodel.StartTaskNodesSet;

public class CNetFixModel {
	public static boolean modelHasErrors(Flex model, StartTaskNodesSet startTaskNodesSet, EndTaskNodesSet endTaskNodesSet, List<String> messageList, List<String> codeList) {
		
		boolean needsFix = false;

		if (startTaskNodesSet.size() == 0) {
			needsFix = true;
			codeList.add("E1");
			messageList
					.add("E1: The provided C-Nets has problems with the start activity. It has no start activity node set.");
			//no solution for having different activities in different sets. (setFlex)
			//solution exists if there is one set which contains the start activity and other sets contain only empty binding.
		} else if (startTaskNodesSet.size() > 1) {
			needsFix = true;
			codeList.add("E1");
			messageList
					.add("E1: The provided C-Nets has problems with the start activity. It has more than one start activity node set.");
			//no solution for having different activities in different sets. (setFlex)
			//solution exists if there is one set which contains the start activity and other sets contain only empty binding.
		} else {
			for (SetFlex setFlex : startTaskNodesSet) {
				if (setFlex.size() != 1) {
					needsFix = true;
					codeList.add("E3");
					messageList
							.add("E3: The provided C-Nets has problems with the start activity. Probably it has more than one start activity or there is no start activity");
					//no solution for having two different activities
					//solution exists if the other activity is an empty biding
					//solution exists: if there is no start activity. We need to find an activity which has an empty input binding.
				} else {
					Collection<FlexEdge<? extends FlexNode, ? extends FlexNode>> inEdges = model.getInEdges(setFlex
							.first());
					if (inEdges.size() > 0) {
						needsFix = true;
						codeList.add("E3");
						messageList
								.add("E3: The provided C-Nets has problems with the start activity. Probably start activity node does not have an empty input binding set!");

					}
				}
			}
		}

		if (endTaskNodesSet.size() == 0) {
			needsFix = true;
			codeList.add("E2");
			messageList
					.add("E2: The provided C-Nets has problems with the end activity. It has no end activity node set.");
			//no solution for having different activities in different sets. (setFlex)
			//solution exists if there is one set which contains the start activity and other sets contain only empty binding.
		} else if (endTaskNodesSet.size() > 1) {
			needsFix = true;
			codeList.add("E2");
			messageList
					.add("E2: The provided C-Nets has problems with the end activity. It has more than one end activity node set.");
			//no solution for having different activities in different sets. (setFlex)
			//solution exists if there is one set which contains the start activity and other sets contain only empty binding.
		} else {

			for (SetFlex setFlex : endTaskNodesSet) {
				if (setFlex.size() != 1) {
					needsFix = true;
					codeList.add("E4");
					messageList
							.add("E4: The provided C-Nets has problems with the end activity. Probably it has more than one end activity or there is no end activity");
					//no solution for having two different activities
					//solution exists if the other activity is an empty biding
					//solution exists: if there is no start activity. We need to find an activity which has an empty output binding.
				} else {
					Collection<FlexEdge<? extends FlexNode, ? extends FlexNode>> outEdges = model.getOutEdges(setFlex
							.first());
					if (outEdges.size() > 0) {
						needsFix = true;
						codeList.add("E4");
						messageList
								.add("E4: The provided C-Nets has problems with the end activity node. Probably end activity node does not have an empty output binding set!");

					}
				}
			}
		}

		boolean controlCheck = false;
		for (FlexNode flexNode : model.getNodes()) {

			for (SetFlex setFlex : flexNode.getInputNodes()) {
				if (setFlex.isEmpty()) {
					codeList.add("E5");
					messageList
							.add("E5: The provided C-Nets has empty bindings due to noisy behavior captured by Causal Net");
					//solution exists by removing an empty biding
					controlCheck = true;
					break;
				}
			}
			if (controlCheck) {
				break;
			}
			for (SetFlex setFlex : flexNode.getOutputNodes()) {
				if (setFlex.isEmpty()) {
					codeList.add("E5");
					messageList
							.add("E5: The provided C-Nets has empty bindings due to noisy behavior captured by Causal Net");
					//solution exists by removing an empty biding
					controlCheck = true;
					break;
				}
			}

			if (controlCheck) {
				break;
			}
		}
		return needsFix || controlCheck;
	}

	public static boolean solveErrors(Flex model, StartTaskNodesSet startTaskNodesSet, EndTaskNodesSet endTaskNodesSet, List<String> analysisResult) {
		boolean isFixSuccessful = true;
		
		if (startTaskNodesSet.size() == 0) {
			isFixSuccessful = false;
			analysisResult.add("<p>There are no start nodes. Please check the correctness of your C-Net!</p>");
			//no solution for having different activities in different sets. (setFlex)
			//solution exists if there is one set which contains the start activity and other sets contain only empty binding.
		} else if (startTaskNodesSet.size() > 1) {
			analysisResult.add("<p>There are multiple start nodes. Please filter the event log to have only one start node!</p>");
			isFixSuccessful = false;
			//no solution for having different activities in different sets. (setFlex)
			//solution exists if there is one set which contains the start activity and other sets contain only empty binding.
		} else {
			for (SetFlex setFlex : startTaskNodesSet) {
				if (setFlex.size() != 1) {
					analysisResult.add("<p>There are multiple start nodes. Please filter the event log to have only one start node!</p>");
					isFixSuccessful = false;
					//no solution for having two different activities
					//solution exists if the other activity is an empty biding
					//solution exists: if there is no start activity. We need to find an activity which has an empty input binding.
				} else {
					Collection<FlexEdge<? extends FlexNode, ? extends FlexNode>> inEdges = model.getInEdges(setFlex
							.first());
					if (inEdges.size() > 0) {
						isFixSuccessful = false;
						analysisResult.add("<p>The start node does not have an empty input binding set. Please filter the event log and add an artificial start event!</p>");

					}
				}
			}
		}
		if (endTaskNodesSet.size() == 0) {
			isFixSuccessful = false;
			analysisResult.add("<p>There are no end nodes. Please check the correctness of your C-Net!</p>");
			//no solution for having different activities in different sets. (setFlex)
			//solution exists if there is one set which contains the start activity and other sets contain only empty binding.
		} else if (endTaskNodesSet.size() > 1) {
			analysisResult.add("<p>There are multiple end nodes. Please filter the event log to have only one end node!</p>");
			isFixSuccessful = false;
			//no solution for having different activities in different sets. (setFlex)
			//solution exists if there is one set which contains the start activity and other sets contain only empty binding.
		} else {
			for (SetFlex setFlex : endTaskNodesSet) {
				if (setFlex.size() != 1) {
					analysisResult.add("<p>There are multiple end nodes. Please filter the event log to have only one start node!</p>");
					isFixSuccessful = false;
					//no solution for having two different activities
					//solution exists if the other activity is an empty biding
					//solution exists: if there is no start activity. We need to find an activity which has an empty input binding.
				} else {
					Collection<FlexEdge<? extends FlexNode, ? extends FlexNode>> outEdges = model.getOutEdges(setFlex
							.first());
					if (outEdges.size() > 0) {
						isFixSuccessful = false;
						analysisResult.add("<p>The end node does not have an empty output binding set. Please filter the event log and add an artificial end event!</p>");

					}
				}

			}
		}

		List<SetFlex> tempSetFlex = new ArrayList<SetFlex>();
		for (FlexNode flexNode : model.getNodes()) {
			tempSetFlex.clear();
			for (SetFlex setFlex : flexNode.getInputNodes()) {
				if (setFlex.isEmpty()) {
					tempSetFlex.add(setFlex);
				}
			}
			flexNode.getInputNodes().removeAll(tempSetFlex);
			tempSetFlex.clear();
			for (SetFlex setFlex : flexNode.getOutputNodes()) {
				if (setFlex.isEmpty()) {
					tempSetFlex.add(setFlex);
				}
			}
			flexNode.getOutputNodes().removeAll(tempSetFlex);
		}

		if (isFixSuccessful) {
			return true;
		} else {
			return false;
		}

	}

}
